/* replace word correspond to dictionary */
/*   version 0.21      2 Nov 1999  Rick  */

/* original OSAKA.X 1989 Katsunari Seki  */


#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#ifndef unchar
#define	unchar	unsigned char
#endif

#ifndef TRUE
#define	TRUE	1
#endif

#ifndef FALSE
#define	FALSE	0
#endif

#define	VERSION	" version 0.21 (origin OSAKA.X Katsunari Seki)\n"
#define	TABLESIZE	1024	/* dictionary look-up table */
#define	BUFSIZE	256

#define	NORMAL	0
#define	BEEP	1
#define	COLOR	2

typedef struct{
	void	*next;			/* pointer to next item */
	int	len;				/* length of string */
	unchar	word;			/* dummy definition */
							/*  old word + new word in real */
} dict;


unchar	*prog;


static void
usage ()
{
	fputs ("usage: ", stderr);
	fputs (prog, stderr);
	fputs (" [-bcV] [-d dictionary] [<files...>]\n", stderr);
	exit (-1);
}

static void
putversion ()
{
	fputs (prog, stderr);
	fputs (VERSION, stderr);
	usage ();
}

static FILE *
openfile (filename)
unchar	*filename;
{
	FILE	*fp;

	fp = fopen (filename, "rt");
	if (!fp) {
		fputs (prog, stderr);
		fputs (": ", stderr);
		fputs (filename, stderr);
		fputs (" cannot open\n", stderr);
	}
	return fp;
}

static void
memerr ()
{
	fputs (prog, stderr);
	fputs (": memory exhausted\n", stderr);
}

static void
killdic (dic)
dict	**dic;
{
	int	i;
	dict	*parent, *child;

	for (i= 0; i < TABLESIZE; i++) {
		if (parent = dic[i]) {
							/* free chained buffer downward */
			do {
				child = parent->next;
				free (parent);
			} while (parent= child);
		}
	}
	free (dic);
}

static unchar *
getword (ptr)
unchar	**ptr;
{
	unchar	*head, *p;

	for (head = *ptr; *head && *head <= ' '; head++);
	for (p = head; *p > ' '; p++);
	*p = '\0';
	*ptr = p;

	return head;
}

static int
adddic (dic, buf)
dict	**dic;
unchar	*buf;
{
	int	len1, len2;
	dict	*newdic;
	dict	*child;
	unchar	*word1, *word2;
	unchar	*ptr = buf;
	int	hash;
								/* get old word */
	word1 = getword (&ptr);
	len1 = ptr - word1;
	if (!len1)	return TRUE;
								/* get new word */
	ptr++;
	word2 = getword (&ptr);
	len2 = ptr - word2;

	newdic = malloc (sizeof(void*) + sizeof(int)
					+ sizeof(unchar) * (len1 + len2 + 2));
	if (newdic){
		newdic->next = NULL;	/* init (no child) */
		newdic->len = len1;
		ptr = &(newdic->word);	/* old word */
		while (*word1)	*ptr++ = *word1++;
		*ptr++ = '\0';
		while (*word2)	*ptr++ = *word2++;
		*ptr = '\0';

		hash = ((newdic->word & 7) << 7)
				| ((*(&(newdic->word) + 1) - 64) & 127);
		child = dic[hash];
		if (child) {			/* if already entried */
			while (child->next)	child = child->next;
			child->next = newdic;	/* add to tail */
		} else {
			dic[hash] = newdic;
		}
		return TRUE;
	} else {
		memerr ();
		return FALSE;
	}
}

static dict **
readdic (filename)
unchar	*filename;
{
	int	i;
	FILE	*dfp;
	dict	**dic;
	unchar	buf[BUFSIZE];

	dic = malloc (sizeof(dict**) * TABLESIZE);
	if (dic) {
		for (i= 0; i < TABLESIZE; i++)	dic[i] = NULL;

		dfp = openfile (filename);
		if (!dfp)	return NULL;

		while (fgets( buf, BUFSIZE, dfp))
			if (!adddic (dic, buf))	break;

		fclose (dfp);
		return dic;
	} else {
		memerr ();
		return NULL;
	}
}

static void
colorput (d)
dict	*d;
{
	fputs ("\x1b[35m", stdout);
	fputs (&(d->word) + d->len + 1, stdout);
	fputc ('(', stdout);
	fputs (&(d->word), stdout);
	fputs (")\x1b[m", stdout);
}

static void
simpleput (d)
dict	*d;
{
	fputs (&(d->word) + d->len + 1, stdout);
}

static void
beepcolorput (d)
dict	*d;
{
	fputc ('\x07', stdout);
	colorput (d);
}

static void
beepput (d)
dict	*d;
{
	fputc ('\x07', stdout);
	simpleput (d);
}

static void
conv (putfunc, dic, ifp)
void	(*putfunc)(dict*);
dict	**dic;
FILE	*ifp;
{
	int	kanji = FALSE;
	unchar	buf[BUFSIZE];

	while (fgets (buf, BUFSIZE, ifp)) {
		unchar	*ptr = buf;
						/* through if rest multi byte code */
		if (kanji) {
			kanji = FALSE;
			fputc (*ptr++, stdout);
		}
		while (*ptr) {
			int	len = 0;
			int	hash;
			dict	*d;
						/* dictionary matching */
			hash = ((ptr[0] & 7) << 7) | ((ptr[1] - 64) & 127);
			if (d = dic[hash]){
				while (d) {
					len = d->len;
					if (!strncmp(&(d->word), ptr, len)) {
						putfunc (d);
						ptr += len;
						break;
					}
					d = d->next;
					len = 0;
				}
			}
			if (!len) {
						/* multi byte check */
#ifdef USE_EUC
				if (*ptr >= 0x80) {
#else					/* SHIFT JIS */
				if ((*ptr >= 0x80 && *ptr <= 0x9f)
						|| (*ptr >= 0xe0 && *ptr <= 0xfc)) {
#endif
					if (ptr[1]) {
						fputc (*ptr++, stdout);
					} else {
						kanji = TRUE;
					}
				}
				fputc (*ptr++, stdout);
			}
		}
	}
}

void
main (argc, argv)
int	argc;
unchar	**argv;
{
	int	i, mode;
	FILE	*ifp;
	void	(*putfunc)(dict*);
	dict	**dic;
	unchar	*ptr, dicname[92];

						/* you can use for filter */
	ifp = stdin;
	mode = NORMAL;
						/* get program name */
	ptr = prog = *argv;
	while (*ptr) {
		if (*ptr == '\\' || *ptr == '/' )
			prog = ptr + 1;
		ptr++;
	}
						/* make default dictionary name */
						/* program name [- extention] + ".dic" */
	strcpy (dicname, *argv);
	for (ptr = dicname; *ptr && *ptr != '.'; ptr++);
	strcpy (ptr, ".dic");

						/* argument & option */
	for (i = 1; i < argc; i++ ) {
		if (*(ptr = argv[i]) == '-') {
			*ptr++ = '\0';
			do {
				switch (*ptr++) {
				case 'V':
					putversion ();
					break;
				case 'b':
					mode |= BEEP;
					break;
				case 'c':
					mode |= COLOR;
					break;
				case 'd':
					if (!*ptr) {
						if (++i < argc) {
							ptr = argv[i];
						} else
							usage ();
					}
					strcpy (dicname, ptr);
					*ptr = '\0';
					break;
				case 'h':
					usage ();
					break;
				default:
					fputs ("illegal option -", stderr );
					fputc (*--ptr, stderr );
					fputc ('\n', stderr );
					usage ();
				}
			} while (*ptr);
		} else {
			ifp = NULL;
		}
	}

	dic = readdic (dicname);
	if (!dic)	exit (-2);

	switch (mode){
	case NORMAL:
		putfunc = simpleput;
		break;
	case BEEP:
		putfunc = beepput;
		break;
	case COLOR:
		putfunc = colorput;
		break;
	case BEEP+COLOR:
		putfunc = beepcolorput;
		break;
	}

	if (ifp) {
		conv (putfunc, dic, ifp);
	} else {
		for (i = 1; i < argc; i++) {
			if (*(argv[i])) {
				ifp = openfile (argv[i]);
				if (ifp) {
					conv (putfunc, dic, ifp);
					fclose (ifp);
				}
			}
		}
	}

	killdic (dic);
}
